/*************************************************************************************/
/*      Copyright 2015 Barcelona Supercomputing Center                               */
/*                                                                                   */
/*      This file is part of the NANOS++ library.                                    */
/*                                                                                   */
/*      NANOS++ is free software: you can redistribute it and/or modify              */
/*      it under the terms of the GNU Lesser General Public License as published by  */
/*      the Free Software Foundation, either version 3 of the License, or            */
/*      (at your option) any later version.                                          */
/*                                                                                   */
/*      NANOS++ is distributed in the hope that it will be useful,                   */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*      GNU Lesser General Public License for more details.                          */
/*                                                                                   */
/*      You should have received a copy of the GNU Lesser General Public License     */
/*      along with NANOS++.  If not, see <http://www.gnu.org/licenses/>.             */
/*************************************************************************************/

#include "bf_sched.hpp"

#include "config.hpp"
#include "schedule.hpp"
#include "plugin.hpp"
#include "pollingservice.hpp"

namespace nanos {
namespace ext {

template < typename BasePolicy = BreadthFirst >
class CommThreadScheduler;

class CommThreadSchedPlugin : public Plugin
{
   public:
   CommThreadSchedPlugin() :
      Plugin( "Communications thread scheduling plugin",1 )
   {
   }
   
   virtual void config ( Config &cfg )
   {
      cfg.setOptionsSection( "CommThreadSched module",
            "Communications thread scheduling module" );

      cfg.registerConfigOption ( "sharing-priority-limit",
            NEW Config::PositiveVar( CommThreadSchedPlugin::_sharingPriorityLimit, 0 ),
            "Minimum priority value to allow"
            "communications thread to run other tasks.");
      cfg.registerEnvOption( "sharing-priority-limit",
            "NX_COMMTHREAD_SHARING_PRIORITY" );
      cfg.registerArgOption( "sharing-priority-limit",
            "comthread-sharing-priority" );
   }
   
   virtual void init();

   static int getSharingPriorityLimit() {
      return _sharingPriorityLimit;
   }

   private:
      static int                     _sharingPriorityLimit;
};

template < typename BasePolicy >
class CommThreadScheduler : public BasePolicy
{
private:
   typedef typename BasePolicy::TeamData BaseTeamData;

   struct TeamData : public BaseTeamData
   {
      WDPool *_communication;
      WDPool *_priority;

      TeamData() :
         BaseTeamData(),
         _communication(NULL),
         _priority(NULL)
      {
         _communication = NEW WDPriorityQueue<>( true /* enableDeviceCounter */, true /* optimise option */ );
         _priority      = NEW WDPriorityQueue<>( true /* enableDeviceCounter */, true /* optimise option */ );
      }

      virtual ~TeamData() {
         if( _communication ) delete _communication;
         if( _priority ) delete _priority;
      }

      WDPool& getCommunicationQueue() {
         return *_communication;
      }

      WDPool& getPriorityQueue() {
         return *_priority;
      }
   };

public:
   CommThreadScheduler() :
      BasePolicy()
   {
   }

   virtual ~CommThreadScheduler()
   {
   }

private:
   virtual size_t getTeamDataSize () const {
      return sizeof(TeamData);
   }
   
   virtual ScheduleTeamData * createTeamData ()
   {
      return NEW TeamData();
   }
   
   virtual void queue ( BaseThread *thread, WD &wd )
   {
      TeamData &tdata = (TeamData &) *thread->getTeam()->getScheduleData();
      BaseThread *target = wd.isTiedTo();
   
      if( target ) {
         target->addNextWD( &wd );
      } else if( wd.getCommFlag() ) {
         tdata.getCommunicationQueue().push_back( &wd );
      } else if( wd.getPriority() >= CommThreadSchedPlugin::getSharingPriorityLimit() ) {
         tdata.getPriorityQueue().push_back( &wd );
      } else {
         BasePolicy::queue(thread, wd);
      }
   }
   
   virtual void queue ( BaseThread** threads, WD** wds, size_t numElems )
   {
      for( ; numElems > 0; --numElems ) {
         queue( *threads, *(*wds) );
         ++wds;
         ++threads;
      }
   }
   
    virtual void atSuccessor( DependableObject &successor, DependableObject &predecessor )
    {
       // Smart priority unsupported
    }
   
   virtual WD* getNextWD( BaseThread *thread )
   {
      bool isCommThread = static_cast<int>(sys.getSMPPlugin()->getNumThreads()) == (thread->getId()+1);
      TeamData &tdata = (TeamData &) *thread->getTeam()->getScheduleData();
   
      WD* result = NULL;
   
      // Check prefetch queue
      if( thread->hasNextWD()  ) {
         return thread->getNextWD();
      } else if( isCommThread ) {
         // Communication thread:
         // take a ready task from the communications task queue when
         // no tasks are present on the prefetch queue
         result = tdata.getCommunicationQueue().pop_front(thread);
         if( result ) {
            // Communication tasks must be executed by comm thread only
            result->tieTo(*thread);
         } else {
            // No communication task available, take one from the
            // priority queue Communication thread does not take tasks
            // from the regular (non priority) queue
            result = tdata.getPriorityQueue().pop_front(thread);
         }
      } else {
         // Regular worker thread:
         // No tasks on prefetch queue, take a ready task from the
         // ready queue as usual.
         result = tdata.getPriorityQueue().pop_front(thread);
         if( !result ) {
            result = BasePolicy::atIdle(thread, false);
         }
      }
      return result;
   }
   
   virtual WD * atBlock ( BaseThread *thread, WD *current )
   {
      // Do not call atIdle directly. We might poll a condition that is blocked by us.
      // As we already checked that condition three times at least, just give up
      // and switch context.
      return getNextWD(thread);
   }
   
   virtual WD * atIdle ( BaseThread *thread, int numSteal = 0 )
   {
      // Callback to registered polling services
      PollingServiceManager::poll();
   
      return getNextWD(thread);
   }
   
   virtual bool testDequeue()
   {
      TeamData &tdata = (TeamData &) *myThread->getTeam()->getScheduleData();
      bool result = BasePolicy::testDequeue() && !tdata.getPriorityQueue().empty();
      return result;
   }
   
   virtual bool usingPriorities() const
   {
      return true;
   }
};

void CommThreadSchedPlugin::init()
{
   sys.setDefaultSchedulePolicy(NEW CommThreadScheduler<>());
}

bool BreadthFirst::_useStack = false;
bool BreadthFirst::_usePriority = true;
bool BreadthFirst::_useSmartPriority = false;

int CommThreadSchedPlugin::_sharingPriorityLimit;

} // namespace ext
} // namespace nanos

DECLARE_PLUGIN("sched-polling",nanos::ext::CommThreadSchedPlugin);