/*************************************************************************************/
/*      Copyright 2015 Barcelona Supercomputing Center                               */
/*                                                                                   */
/*      This file is part of the NANOS++ library.                                    */
/*                                                                                   */
/*      NANOS++ is free software: you can redistribute it and/or modify              */
/*      it under the terms of the GNU Lesser General Public License as published by  */
/*      the Free Software Foundation, either version 3 of the License, or            */
/*      (at your option) any later version.                                          */
/*                                                                                   */
/*      NANOS++ is distributed in the hope that it will be useful,                   */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*      GNU Lesser General Public License for more details.                          */
/*                                                                                   */
/*      You should have received a copy of the GNU Lesser General Public License     */
/*      along with NANOS++.  If not, see <http://www.gnu.org/licenses/>.             */
/*************************************************************************************/

#include "config_decl.hpp"
#include "config.hpp"
#include "schedule.hpp"
#include "plugin.hpp"
#include "priority_queue.hpp"
#include "pollingservice.hpp"

#include <unordered_set>

namespace nanos {
   typedef Config::ListOption<std::list<int>, Config::IntegerHelpFormat> PriorityVarList;

namespace ext {
      class CommThreadSchedPlugin : public Plugin
      {

         public:
            CommThreadSchedPlugin() :
               Plugin( "Communications thread scheduling plugin",1 ) {}

            virtual ~CommThreadSchedPlugin()
            {
#ifdef NANOS_INSTRUMENTATION_ENABLED
               if( !_dumpFilename.empty() ) {
                  std::fstream dump(_dumpFilename, std::ios_base::out);
                  for( int priority : _usedPriorities ) {
                     dump << priority << " ";
                  }

                  message0( "Task priority values were stored in file "
                            << _dumpFilename
                            << std::endl );
               }
#endif
            }

            virtual void config ( Config &cfg )
            {
               cfg.setOptionsSection( "CommThreadSched module",
                     "Communications thread scheduling module" );

               cfg.registerConfigOption ( "sharing-priority-limit",
                     NEW Config::PositiveVar( CommThreadSchedPlugin::_sharingPriorityLimit, 0 ),
                     "Minimum priority value to allow"
                     "communications thread to run other tasks.");
               cfg.registerEnvOption( "sharing-priority-limit",
                     "NX_COMMTHREAD_SHARING_PRIORITY" );
               cfg.registerArgOption( "sharing-priority-limit",
                     "comthread-sharing-priority" );

               cfg.registerConfigOption ( "priority-list",
                     NEW PriorityVarList( CommThreadSchedPlugin::_configuredPriorities ),
                     "Space separated integer list of priorities.");
               cfg.registerEnvOption( "priority-list",
                     "NX_COMMTHREAD_PRIORITIES" );
               cfg.registerArgOption( "priority-list",
                     "comthread-priorities" );
#ifdef NANOS_INSTRUMENTATION_ENABLED
               cfg.registerConfigOption ( "dump-priorities",
                     NEW Config::StringVar( CommThreadSchedPlugin::_dumpFilename ),
                     "Name of the file to dump used priorities.");
               cfg.registerEnvOption( "dump-priorities",
                     "NX_COMMTHREAD_DUMP_FILE" );
               cfg.registerArgOption( "dump-priorities",
                     "comthread-dump-file" );
#endif
            }

            virtual void init();

            static int getSharingPriorityLimit() { return _sharingPriorityLimit; }
            static const std::list<int>& getConfiguredPriorities() { return _configuredPriorities; }

#ifdef NANOS_INSTRUMENTATION_ENABLED
            static std::unordered_set<int>& getUsedPriorities() { return _usedPriorities; }
            static const std::string&       getDumpFilename() { return _dumpFilename; }
#endif

         private:
            static int                     _sharingPriorityLimit;
            static std::list<int>          _configuredPriorities;
#ifdef NANOS_INSTRUMENTATION_ENABLED
            static std::unordered_set<int> _usedPriorities;
            static std::string             _dumpFilename;
#endif
      };

      class CommThreadBF : public SchedulePolicy
      {
         private:
            class PriorityReadyQueue {
               BucketPriorityQueue<int,WorkDescriptor*> _queue;
#ifdef NANOS_INSTRUMENTATION_ENABLED
               Atomic<unsigned int>                     _numReadyTasks;
#endif

               public:
#ifdef NANOS_INSTRUMENTATION_ENABLED
               template< typename InputIt >
                  PriorityReadyQueue( InputIt begin, InputIt end ) :
                     _queue(begin,end),
                     _numReadyTasks(0)
               {
               }
#else
               template< typename InputIt >
                  PriorityReadyQueue( InputIt begin, InputIt end ) :
                     _queue(begin,end)
               {
               }
#endif

               void insert( WD& task ) {
                  NANOS_INSTRUMENT( ++_numReadyTasks );
                  NANOS_INSTRUMENT( static nanos_event_key_t key = sys.getInstrumentation()->getInstrumentationDictionary()->getEventKey("num-ready") );
                  NANOS_INSTRUMENT( nanos_event_value_t nb =  static_cast<nanos_event_value_t>(_numReadyTasks) );
                  NANOS_INSTRUMENT( sys.getInstrumentation()->raisePointEvents(1, &key, &nb) );

                  _queue.insert( task.getPriority(), &task );
               }

               template< typename InputIt >
                  void insert( InputIt begin, InputIt end ) {
                     struct GetWDPriority {
                        int operator()( const WD* element ) const {
                           return element->getPriority();
                        }
                     };

                     NANOS_INSTRUMENT( _numReadyTasks += std::distance(begin,end) );
                     NANOS_INSTRUMENT( static nanos_event_key_t key = sys.getInstrumentation()->getInstrumentationDictionary()->getEventKey("num-ready") );
                     NANOS_INSTRUMENT( nanos_event_value_t nb =  static_cast<nanos_event_value_t>(_numReadyTasks) );
                     NANOS_INSTRUMENT( sys.getInstrumentation()->raisePointEvents(1, &key, &nb) );

                     // Assumes all the threads belong to the same team
                     _queue.insert_all( begin, end, GetWDPriority() );
                  }

               Optional<WD*> extract() {
#ifdef NANOS_INSTRUMENTATION_ENABLED
                  Optional<WD*> result = _queue.extract();
                  if( result.has_value() ) {
                     --_numReadyTasks;
                     static nanos_event_key_t key = sys.getInstrumentation()->getInstrumentationDictionary()->getEventKey("num-ready");
                     nanos_event_value_t nb =  static_cast<nanos_event_value_t>(_numReadyTasks);
                     sys.getInstrumentation()->raisePointEvents(1, &key, &nb);
                  }
                  return result;
#else
                  return _queue.extract();
#endif
               }

               bool empty() const {
                  return _queue.empty();
               }
            };

            struct IsWorkerPriority {
               int upperbound;

               IsWorkerPriority() :
                  upperbound(CommThreadSchedPlugin::getSharingPriorityLimit())
               {}

               bool operator()( int priority ) {
                  return priority < upperbound;
               } 
            };

            typedef PriorityReadyQueue CommTaskReadyQueue;

            struct TeamData : public ScheduleTeamData
         {
            IsWorkerPriority   _isWorkerPriority;
            PriorityReadyQueue _worker;
            PriorityReadyQueue _shared;
            CommTaskReadyQueue _communication;

            // Specify which priority values to use as buckets
            // begin, end are iterators of a list of priorities
            // Assuming priorities are sorter from greater to lower
            template< typename InputIt >
               TeamData( InputIt begin, InputIt end ) :
                  _worker(std::find_if(begin,end,_isWorkerPriority),end),
                  _shared(begin,end),
                  _communication(begin,end)
            {
            }

            PriorityReadyQueue& getWorkerQueue() {
               return _worker;
            }

            PriorityReadyQueue& getSharedQueue() {
               return _shared;
            }

            CommTaskReadyQueue& getCommunicationQueue() {
               return _communication;
            }
         };

         public:
            CommThreadBF() :
               SchedulePolicy("Polling conditions and breadth First")
         {
         }

            virtual ~CommThreadBF () {}

         private:

            virtual size_t getTeamDataSize () const { return sizeof(TeamData); }
            virtual size_t getThreadDataSize () const { return 0; }

            virtual ScheduleTeamData * createTeamData ()
            {
               const std::list<int>& priorities = CommThreadSchedPlugin::getConfiguredPriorities();
               if( priorities.empty() ) {
                  int default_priorities[] = {0};
                  return NEW TeamData( std::begin(default_priorities), std::end(default_priorities) );
               } else {
                  return NEW TeamData( priorities.begin(), priorities.end() );
               }
            }

            virtual ScheduleThreadData * createThreadData ()
            {
               return 0;
            }

            virtual void queue ( BaseThread *thread, WD &wd )
            {
               TeamData &tdata = (TeamData &) *thread->getTeam()->getScheduleData();
               BaseThread *target = wd.isTiedTo();
#ifdef NANOS_INSTRUMENTATION_ENABLED
               if( !CommThreadSchedPlugin::getDumpFilename().empty() ) {
                  CommThreadSchedPlugin::getUsedPriorities().insert(wd.getPriority());
               }
#endif

               if( target ) {
                  target->addNextWD( &wd );
               } else if( wd.getCommFlag() ) {
                  tdata.getCommunicationQueue().insert( wd );
               } else if( tdata._isWorkerPriority(wd.getPriority()) ) {
                  tdata.getWorkerQueue().insert( wd );
               } else {
                  tdata.getSharedQueue().insert( wd );
               }
            }

            virtual void queue ( BaseThread** threads, WD** wds, size_t numElems )
            {
               for( ; numElems > 0; --numElems ) {
                  queue( *threads, *(*wds) );
                  ++wds;
                  ++threads;
               }
            }

            /*! This scheduling policy supports all WDs, no restrictions. */
            bool isValidForBatch ( const WD * wd ) const {
               return true;
            }

            void atSuccessor( DependableObject &successor, DependableObject &predecessor )
            {
               // Smart priority unsupported
            }

            virtual WD *atSubmit ( BaseThread *thread, WD &newWD )
            {
               queue( thread, newWD );
               return 0;
            }

            WD* getNextWD( BaseThread *thread )
            {
               bool isCommThread = static_cast<int>(sys.getSMPPlugin()->getNumThreads()) == (thread->getId()+1);
               TeamData &tdata = (TeamData &) *thread->getTeam()->getScheduleData();

               Optional<WD*> result;
               // Check prefetch queue after service poll. A task might have been woken up.
               if( thread->hasNextWD()  ) {
                  return thread->getNextWD();
               } else if( isCommThread ) {
                  // Communication thread:
                  // take a ready task from the communications task queue when
                  // no tasks are present on the prefetch queue
                  result = tdata.getCommunicationQueue().extract();
                  if( result.has_value() ) {
                     // Communication tasks must be executed by comm thread only
                     result.value()->tieTo(*thread);
                  } else {
                     // No communication task available, take one from the
                     // priority queue Communication thread does not take tasks
                     // from the regular (non priority) queue
                     result = tdata.getSharedQueue().extract();
                  }
               } else {
                  // Regular worker thread:
                  // No tasks on prefetch queue, take a ready task from the
                  // ready queue as usual.
                  result = tdata.getSharedQueue().extract();
                  if( !result.has_value() ) {
                     result = tdata.getWorkerQueue().extract();
                  }
               }
               return result.has_value()? result.value() : NULL;
            }

            WD * atBlock ( BaseThread *thread, WD *current )
            {
               // Do not call atIdle directly. We might poll a condition that is blocked by us.
               // As we already checked that condition three times at least, just give up
               // and switch context.
               return getNextWD(thread);
            }

            WD * atIdle ( BaseThread *thread, int numSteal = 0 )
            {
               // Callback to registered polling services
               PollingServiceManager::poll();

               return getNextWD(thread);
            }

            WD * atPrefetch ( BaseThread *thread, WD &current )
            {
               return atIdle(thread);
            }

            WD * atBeforeExit ( BaseThread *thread, WD &current, bool schedule )
            {
               return NULL;
            }

            virtual void atShutdown()
            {
            }

            bool testDequeue()
            {
               TeamData &tdata = (TeamData &) *myThread->getTeam()->getScheduleData();
               bool result = !( tdata.getWorkerQueue().empty() && tdata.getSharedQueue().empty() );
               //bool result = true;
               return result;
            }

            bool usingPriorities() const
            {
               return true;
            }
      };

      void CommThreadSchedPlugin::init() {
               std::greater<int> highToLowPriorityCompare;
               _configuredPriorities.sort( highToLowPriorityCompare );

                  std::cout << "Configured priorities: ";
                  for( int prio : _configuredPriorities ) {
                     std::cout << prio << " ";
                  }
                  std::cout << std::endl;

               sys.setDefaultSchedulePolicy(NEW CommThreadBF());
      }

      int                     CommThreadSchedPlugin::_sharingPriorityLimit;
      std::list<int>          CommThreadSchedPlugin::_configuredPriorities;
#ifdef NANOS_INSTRUMENTATION_ENABLED
      std::unordered_set<int> CommThreadSchedPlugin::_usedPriorities;
      std::string             CommThreadSchedPlugin::_dumpFilename;
#endif
   }
}

DECLARE_PLUGIN("sched-commthread-tune",nanos::ext::CommThreadSchedPlugin);
